local AV = celestia:newvector(0,0,-1) -- forward
local UP = celestia:newvector(0,1,0)
local GA = celestia:newvector(1,0,0)
local km = 9460730.4725808
local StarSystem, model3D
local reactivity
local dx, dy, dz, defx, defy, defz
local wobble
local angle, dangle = 0, 1/50

local frame, positionframe, orientationframe
local oldorient, oldposition, cockpitorient
local moteur, gouv

function suivi_orbit(t)
   local orbit = {};
   orbit.boundingRadius = 1

   StarSystem = t.StarSystem
   model3D = t.Model3D
   dx = t.dx /1000
   dy = t.dy /1000
   dz = t.dz /1000
   defx, defy, defz = dx, dy, dz
   wobble = t.wobble
   
   reactivity = math.min(math.max(t.reactivity, 0.01),1)
   
   function orbit:position(tjd)
		moteur = moteur or celestia:find("Sol/"..model3D.."_translation_gear") 
		wobble = wobble or celestia:find(StarSystem.."/"..model3D):radius()
		local centre = celestia:find(StarSystem)
		frame = celestia:newframe("ecliptic",centre)
		local obs = celestia:getobserver()
		local obsframe = obs:getframe()
		local obsposition = obs:getposition()
		local orient = obs:getorientation()
		local pos
		if moteur:radius() >= 1 then -- reinit position
			if moteur:radius() == 1 then
				dx, dy, dz = defx, defy, defz
			elseif moteur:radius() == 2 then
				dx, dy, dz = 0, 0, 0
			end
			oldposition, cockpitorient = nil, nil
			moteur:setradius(0.1)
		end
		if moteur:visible()  or not oldposition then
			positionframe = obsframe
			pos = obsposition
			local dist = math.sqrt(dx*dx + dy*dy + dz*dz)
			
			local ddx, ddy, ddz = dx, dy, dz
			if dist ~=0 then
				-- atlantis only
				if dist > wobble then 
					angle = angle + dangle
					ddx = ddx + math.sin(angle*1.234567) * wobble *0.02
					ddy = ddy + math.sin(angle*2.7791) * wobble *0.005
					ddz = ddz + math.sin(angle+1.1234) * wobble * 0.01
					dist = math.sqrt(ddx*ddx + ddy*ddy + ddz*ddz)
				end
				---- fin atlantis
				local dir = (AV * ddz) + (UP *ddy) +(GA * ddx); dir = dir:normalize()
				local vdir
				if cockpitorient then
					vdir = cockpitorient:transform(dir) * (dist/km)
				else
					vdir = orient:transform(dir) * (dist/km)
				end
				pos = pos:addvector(vdir) --cockpit position
			end
			oldposition= positionframe:to(pos) 
		else
			pos = positionframe:from(oldposition)
			local depli = obsposition:vectorto(pos)
			orient = cockpitorient or orient
			local imag, real = orient:imag(), orient:real()
			local invrot = celestia:newrotation(-real, imag.x, imag.y, imag.z)
			local depl = invrot:transform(depli)
			dx = depl.x * km
			dy = depl.y * km
			dz = -depl.z * km
		end
		pos = frame:to(pos)
		return pos:getx() * km , -pos:getz() * km, pos:gety() * km
    end

    return orbit
end 

function suivi_attitude(t)
	local attitude = {}
	
	function attitude:orientation(tjd)
		gouv = gouv or celestia:find("Sol/"..model3D.."_rotation_gear")
		local obs=celestia:getobserver()
		local obsframe = obs:getframe()
		local orient = obs:getorientation()
		if gouv:visible() or not cockpitorient then
			orientationframe = obsframe
			if cockpitorient then
				orient=cockpitorient:slerp(orient,reactivity)
			end
			cockpitorient = orient
			local garagetime = celestia:gettime()
			oldorient = orientationframe:to(orient,garagetime)
		else
			cockpitorient = orientationframe:from(oldorient)
		end
		orient = frame:to(cockpitorient)
		local vector = orient:imag()
		local w = orient:real()
		return w, vector.x, vector.y, vector.z
	end

	return attitude
end
