local AV = celestia:newvector(0,0,-1) -- forward
local UP = celestia:newvector(0,1,0)
local GA = celestia:newvector(1,0,0)
local km = 9460730.4725808
local StarSystem, model3D
local reactivity
local dx, dy, dz, defx, defy, defz
local cockpitspeed, ckspeed
local standardmode = true
local frame, centre
local positionframe, orientationframe, cockpitorient
local oldorient, oldposition
local moteur, gouv

function suivi_orbit(t)
   local orbit = {};
   orbit.boundingRadius = 1

   StarSystem = t.StarSystem
   model3D = t.Model3D
   dx = t.dx /1000
   dy = t.dy /1000
   dz = t.dz /1000
   defx, defy, defz = dx, dy, dz
   
   reactivity = math.min(math.max(t.reactivity, 0.01),1)
   
   function orbit:position(tjd)
		moteur = moteur or celestia:find("Sol/"..model3D.."_translation_gear") 
		centre = centre or celestia:find(StarSystem)
		frame = frame or celestia:newframe("ecliptic",centre)
		local obs = celestia:getobserver()
		local obsframe = obs:getframe()
		local obsposition = obs:getposition()
		local pos
		if moteur:radius() >= 1 then -- reinit position
			if moteur:radius() == 1 then dx, dy, dz = defx, defy, defz
				elseif moteur:radius() == 2 then dx, dy, dz = 0, 0, 0
			end
			oldposition, cockpitorient = nil, nil
			moteur:setradius(0.1)
		end
		cockpitorient = cockpitorient or obs:getorientation()
		if standardmode  or not oldposition then
			positionframe = obsframe
			local dir = (AV * dz) + (UP *dy) +(GA * dx)
			local vdir = cockpitorient:transform(dir) * (1/km)
			pos = obsposition:addvector(vdir) --cockpit position
			
			if oldposition then
				cockpitspeed = positionframe:to(pos) - oldposition
				ckspeed = math.sqrt(cockpitspeed.x * cockpitspeed.x + cockpitspeed.y * cockpitspeed.y + cockpitspeed.z * cockpitspeed.z) > 1e-15
			end
			standardmode = moteur:visible()
		end	
		if not standardmode then	
			pos = positionframe:from(oldposition + cockpitspeed )
			if moteur:visible() then
				local depli = obsposition:vectorto(pos)
				local imag, real = cockpitorient:imag(), cockpitorient:real()
				local invrot = celestia:newrotation(-real, imag.x, imag.y, imag.z)
				local depl = invrot:transform(depli)
				dx = depl.x * km
				dy = depl.y * km
				dz = -depl.z * km
			elseif ckspeed and obsframe:getrefobject():name() ~= model3D then
				obsposition = positionframe:to(obsposition) + cockpitspeed
				obsposition = positionframe:from(obsposition)
				obs:setposition(obsposition)
			end
			standardmode = moteur:visible()		
		end

		oldposition = positionframe:to(pos)
		pos = frame:to(pos)
		return pos:getx() * km , -pos:getz() * km, pos:gety() * km
    end

    return orbit
end 

function suivi_attitude(t)
	local attitude = {}
	
	function attitude:orientation(tjd)
		gouv = gouv or celestia:find("Sol/"..model3D.."_rotation_gear")
		local obs=celestia:getobserver()
		local obsframe = obs:getframe()
		local orient = obs:getorientation()
		if gouv:visible() or not cockpitorient then
			orientationframe = obsframe
			if cockpitorient then
				orient=cockpitorient:slerp(orient,reactivity)
			end
			cockpitorient = orient
			oldorient = orientationframe:to(orient)
		else
			cockpitorient = orientationframe:from(oldorient)
		end
		orient = frame:to(cockpitorient)
		local vector = orient:imag()
		local w = orient:real()
		return w, vector.x, vector.y, vector.z
	end

	return attitude
end
