-- cockpits V3D

if not pcall( function() require ("cockpits_config") end ) then
	celestia:log("--cockpits_config error -- cockpits not loaded")
	exiting=zhorutbvigl[0] -- no need to load the textures
end

if  lua_edu_tools and lang ~= "en"
	then require "locale" 
	else require "cockpits_locale"
end
if pcall( function() require ("locale_cockpits_"..lang) end ) then
	for engl, transl in pairs(cockpits_loc_str) do
		loc_str[engl]=transl
	end
end
require "extras/cockpits/utils/pCXBox"
require "extras/cockpits/utils/_textlayout"

-- constants
local degree = 180/math.pi
local km  = 9460730.4725808 -- mly
local m = km*1000
local km_c = 299792.458; -- km / s
local km_au = 149597870.7
local c_au = km_au/km_c --
local au_ly = 10^6*km /km_c / c_au

local cbutextoff =    {0.8, 0.9, 0.9, 0.8}    -- unactivated button text color
local ctext =     {1, 1, 1, 0.9}    -- main text color
local csetbofill =    {0.05, 0.1 , 0.1, 0.9}    -- setting box filling color
local celestiacolortext = {0.7, 0.7, 1, 1} -- same color than celestia display
local speedontext = {0.4, 1, 0, 1} -- 
local speedofftext = {1,0.15,0,1}
local speedbackground = {0,0,0,0.4}

local cockpitdir = "../panels/" -- relative to the directory of cockpitsBox.lua
local width, height = celestia:getscreendimension()
local oldwidth, oldheight = 0,0

local btnrotTexture = {}
local btnmanTexture = {}
for i=0,14 do 
	if i<5 then btnrotTexture[i]=celestia:loadtexture("btnrot"..i..".png") end
	btnmanTexture[i]=celestia:loadtexture("btndir"..i..".png") 
end

local bug = celestia:loadtexture("BUG.png")

local btnsize = btnsize or 20
local Vmg = Vmg or 10
local becart = btnsize / 4

-- cp = control panel for rotations
local cpsize = btnsize*3+2*becart
local cpX, cpY = (width - cpsize)/2, Vmg

-- mp = manoeuvre panel
local htexte = 20 -- text height at the top of the box
local mpsize = btnsize*7+20*becart
local mpheight = cpsize + 2*becart
local mpX, mpY = cpX+cpsize+becart, cpY

--exp = extracontrol panel
local expsize = 130
local expheight = 9 * 19 -- 9 lignes de 19 pixels
local expX = mpX + mpsize + becart
local expY = Vmg
local showextra = false

-- extra picture
local pictTex = nil

-- choice menu
local cockpits = {}
for i=1, # known_cockpits_files do
	local fic = io.open("extras/cockpits/panels/"..known_cockpits_files[i]..".png","r")
	if not fic then
		fic = io.open("extras/cockpits/panels/"..known_cockpits_files[i].."/"..known_cockpits_files[i]..".lua","r")
	end
	if not fic then
		fic = io.open("extras/cockpits/panels/"..known_cockpits_files[i].."/"..known_cockpits_files[i].."-1.png","r")
	end
	if fic then
		table.insert(cockpits, known_cockpits_files[i])
		fic:close()
	end
end
local ckchoiceWidth = 180
local ckchoiceHeight = #cockpits * 20 + 60
local menuRightpos = 180
if lua_edu_tools then menuRightpos = boxWidth +10 end
local ckchoiceMg = 10

local cockpitsdata ={}
local ck = 0
local panel = 0
local newpanel = false
local multipanel = false
local nbpanel = 0
local loading = false

local obs=celestia:getobserver()
local frame = obs:getframe()
local refobj = frame:getrefobject()
local celestiaspeed = obs:getspeed()
local savecelspeed = celestiaspeed
local fov=obs:getfov()

-- direction and speed management
local UP = celestia:newvector(0,1,0)
local DN = UP*-1
local AV = celestia:newvector(0,0,-1) -- forward
local AR = AV*-1
local GA = celestia:newvector(1,0,0)  -- left
local DR = GA*-1
local direction={DR,GA,UP,DN,AV,AR,nil,DN,UP,DR,GA,AV,AR}
local direct, inverse = 1, -1
local vv, oldvv, savevv = nil, AV, nil  -- speed vector
local celvv = nil -- celestia speed vector
local impulse = 10 ^ (1/5) -- multiply or divide the speed by 10 in 5 key strokes
local power = 1
local react = 0.5 -- nb of seconds to reach powercmd
local powercmd = 1 -- engine speed
local deltapower = 0 -- acceleration / tic
local acceltime = 0
local scripttime = celestia:getscripttime()
local perf = 1/30
local ckspeed = 0 -- speed to display if controlled by the cockpit
local inputdistcolor = {0.1,0.4,0,0.4}
local distance = math.huge
local remainingdistance = distance
local tabunitdist = { {_"m",  1/m}, {_"km", 1/km}, {_"au", km_au / km} , {_"ly", 10^6 } }
local unitidx = 2 -- km by default

--rotation management
local defaultrot = default_rotation or 30
local inputrotcolor = {0,0.4,0.4,0.4}
local rotspeedcoloroff ={0.4,0,0.7,0.4}
local rotspeedcoloron = {0,0.4,0.7,0.4}
local rotationspeed = rotation_speed or 60 -- degree/s
local rotimpulse = 1.4
local deltarot = 0
local oldrotangle 
local oldangleText
local nbrot = 0
local axerot = UP

-- multi panel management
local panelratio, cockpitfov = 0,0
local rearpanel = 0
local frontpanel = 1
local backview = false
local zoomcoef = nil -- zoomcoef =  tan(horizontal_fov /2);   nil=>no zoom management
local azim, elev = 0, 0
local deplX, deplY, deplZ = 0,0,0

-- additionnal visor
local visorsize=visorsize or 100
local visor_active = false
local displayspeed = true

-- extra info
local lspace = normalfont:getheight()+1
local dx, dy, dz, drotX, drotY, drotZ = 0,0,0,0,0,0

-- navaid
local tracked=nil -- no track object selected
local aheadposition = nil
local fracrot=1

--3D cockptis
local cockpit3D = false
local controled3Dcolor = {0.2, 1, 0.2, 0.3}
local free3Dcolor = {0.5, 0.2, 0, 0.8}
local model, model3D, SolarSystem -- model3D=name, model=object
local engine, steering
local cockpitorientation = obs:getorientation()
local oldcockpit3Dselection = celestia:getselection()
local cockpitvector, cockpitspeed


----------------------------------------------
-- Set up and Draw the boxes
----------------------------------------------
cockpitsBox = CXBox:new()
:init(0, 0, 0, 0)
:movable(false)
:attach(screenBox, width, height, 0, 0);

if lua_edu_tools then
	cockpitsCheck = CXBox:new()
    :init(0, 0, 0, 0)
    :bordercolor(cbubordoff)
    :textfont(normalfont)
    :textcolor(cbutextoff)
    :textpos("center")
    :movetext(0, 8)
    :text(_("Set a cockpit"))
    :movable(false)
    :active(true)
    :attach(cockpitsBox, 8, 4, 8, 3)
end

cockpitsSurvey = pCXBox:new()
:init(0, 0, 0, 0)
:movable(false)
:attach(screenBox, width, height, 0, 0);

cockpitsFrame = pCXBox:new()
:init(0, 0, 0, 0)
:movable(false)
:active(false)
:clickable(true)
:visible(false)
:attach(screenBox,0,0,0,0)

ckpictureFrame = pCXBox:new()
:init(0, 0, 0, 0)
:movable(false)
:active(false)
:clickable(true)
:visible(false)
:attach(cockpitsFrame,0,0,0,0)

-- cockpit rotation panel
rotationpanelFrame = pCXBox:new()
:init(0, 0, 0, 0)
:bordercolor({0.8, 0.8, 0.8, 0.8})
:fillcolor({0,0,0,0.6})
:movable(true)
:active(false)
:clickable(false)
:visible(false)
:attach(screenBox, cpX,cpY,width -cpX - cpsize, height - cpY - cpsize)

btncockpitrotFrame ={}
local btnrotposition = function(x)
	local mg,mb,md,mh
	mg=btnsize + becart; mb = mg -- (x==0)
	if x==1 then mg = mg - btnsize - becart
		elseif x==2 then mg = mg + btnsize + becart
		elseif x==3 then mb = mb + btnsize + becart
		elseif x==4 then mb = 0
	end
	md = cpsize - mg - btnsize
	mh = cpsize - mb - btnsize
	return mg, mb, md, mh
end

local createbtnrot = function (x)
	return pCXBox:new()
		:init(0, 0, 0, 0)
		:fillimage(btnrotTexture[x])
		:movable(false)
		:active(true)
		:clickable(false)
		:visible(false)
		:attach(rotationpanelFrame, btnrotposition(x))
end
for i=0,4 do
	btncockpitrotFrame[i]=createbtnrot(i)
	btncockpitrotFrame[i].panel = 0
end

-- manoeuvre panel
manoeuvrepanelFrame = pCXBox:new()
:init(0, 0, 0, 0)
:bordercolor({0.8, 0.8, 0.8, 0.8})
:fillcolor({0,0,0,0.8})
:movable(true)
:active(false)
:clickable(false)
:visible(false)
:attach(screenBox, mpX,mpY,width -mpX - mpsize, height - mpY - mpheight-htexte)

btncockpitmanFrame ={}
local btnmanposition = function(x)
	local mg,mb,md,mh
	mg=5*btnsize + 16*becart; mb = btnsize + 2*becart -- (x==0)
	if x==1 then mg = 4*btnsize + 15*becart
		elseif x==2 then mg = 6*btnsize + 17*becart
		elseif x==3 then mb = mb + btnsize + becart
		elseif x==4 then mb = becart
		elseif x==5 then mg = 10*becart + 3*btnsize; mb = mb + btnsize + becart
		elseif x==6 then mg = 10*becart + 3*btnsize; mb = becart
		elseif x==7 then mg = 10*becart + 3*btnsize;
		
		elseif x==8 then mg = 3*becart
		elseif x==9 then mg = 2*btnsize + 5*becart
		elseif x==10 then mg = btnsize + 4*becart; mb = mb + btnsize + becart
		elseif x==11 then mg = btnsize + 4*becart; mb = becart
		elseif x==12 then mg = 3*becart; mb = mb + btnsize + becart
		elseif x==13 then mg = 2*btnsize + 5*becart; mb = mb + btnsize + becart
		elseif x==14 then mg = btnsize + 4*becart
	end
	md = mpsize - mg - btnsize
	mh = mpheight - mb - btnsize + htexte
	return mg, mb, md, mh
end

local createbtnman = function (x)
	return pCXBox:new()
		:init(0, 0, 0, 0)
		:fillimage(btnmanTexture[x])
		:movable(false)
		:active(true)
		:clickable(false)
		:visible(true)
		:attach(manoeuvrepanelFrame, btnmanposition(x))
end
for i=0,14 do
	btncockpitmanFrame[i]=createbtnman(i)
	local j=i; -- if i==7 then j=0 end -- same function for 0 and 7
	btncockpitmanFrame[i].btn = j
end

	
-- rotation control buttons
local fenwidth = (btnsize*3+becart*8)/2

fenangle = pCXBox:new()
	:init(0, 0, 0, 0)
	:bordercolor({0.8, 0.8, 0.8, 0.8})
	:fillcolor(inputrotcolor)
	:textfont(normalfont)
	:textcolor(ctext)
	:textpos("center")
	:text(tostring(defaultrot))
	:movetext(0, 4)
	:movable(false)
	:active(true)
	:editable(true)
	:clickable(false)
	:visible(true)
	:attach(manoeuvrepanelFrame, 0, mpheight, mpsize - fenwidth,0)

fenspeedrot = pCXBox:new()
	:init(0, 0, 0, 0)
	:bordercolor({0.8, 0.8, 0.8, 0.8})
	:fillcolor(rotspeedcoloroff)
	:textfont(normalfont)
	:textcolor(ctext)
	:textpos("center")
	:text(tostring(rotationspeed))
	:movetext(0, 4)
	:movable(false)
	:clickable(false)
	:visible(true)
	:attach(manoeuvrepanelFrame, fenwidth, mpheight, mpsize - fenwidth*2,0)

fenwidth = fenwidth / 2
decspeedrot = pCXBox:new()
	:init(0, 0, 0, 0)
	:movable(false)
	:active(true)
	:clickable(false)
	:visible(true)
	:attach(fenspeedrot, 0, 0, fenwidth,0)
	
incspeedrot = pCXBox:new()
	:init(0, 0, 0, 0)
	:movable(false)
	:active(true)
	:clickable(false)
	:visible(true)
	:attach(fenspeedrot, fenwidth, 0, 0, 0)
	
-- translation control buttons
fenwidth = (btnsize*3+becart*8)/2

power1xbutton = pCXBox:new()
	:init(0, 0, 0, 0)
	:bordercolor({0.8, 0.8, 0.8, 0.8})
	:textfont(normalfont)
	:textcolor(ctext)
	:movetext(0, 4)
	:textpos("center")
	:text("1x")
	:movable(false)
	:active(true)
	:clickable(false)
	:visible(true)
	:attach(manoeuvrepanelFrame, btnsize*3+becart*8, mpheight, btnsize*3.5+becart*10,0)
	
nominalpowerbutton = pCXBox:new()
	:init(0, 0, 0, 0)
	:bordercolor({0.8, 0.8, 0.8, 0.8})
	:fillcolor({1,0,0,0.25})
	:textfont(normalfont)
	:textcolor(ctext)
	:movetext(0, 0)
	:textpos("center")
	:movable(false)
	:active(true)
	:clickable(false)
	:visible(true)
	:attach(manoeuvrepanelFrame, btnsize*3.5+becart*10, mpheight, fenwidth+btnsize,0)

distancebox = pCXBox:new()
	:init(0, 0, 0, 0)
	:bordercolor({0.8, 0.8, 0.8, 0.8})
	:fillcolor(inputdistcolor)
	:textfont(normalfont)
	:textcolor(ctext)
	:textpos("center")
	:text(">>>")
	:movetext(0, 4)
	:movable(false)
	:active(true)
	:editable(true)
	:clickable(false)
	:visible(true)
	:attach(manoeuvrepanelFrame, fenwidth*3 + btnsize, mpheight, btnsize,0)
	
distanceunitbutton = pCXBox:new()
	:init(0, 0, 0, 0)
	:bordercolor({0.8, 0.8, 0.8, 0.8})
	:fillcolor(inputdistcolor)
	:textfont(normalfont)
	:textcolor(ctext)
	:movetext(0, 4)
	:textpos("center")
	:text(tabunitdist[unitidx][1])
	:movable(false)
	:active(true)
	:clickable(false)
	:visible(true)
	:attach(manoeuvrepanelFrame, fenwidth*4+btnsize, mpheight,0,0)

fenwidth = fenwidth / 2

slowbutton = pCXBox:new()
	:init(0, 0, 0, 0)
	:movable(false)
	:active(true)
	:clickable(false)
	:visible(true)
	:attach(nominalpowerbutton, 0, 0, fenwidth,0)
	
accelbutton = pCXBox:new()
	:init(0, 0, 0, 0)
	:movable(false)
	:active(true)
	:clickable(false)
	:visible(true)
	:attach(nominalpowerbutton, fenwidth, 0, 0, 0)
	
-----

speeddisplaybutton = pCXBox:new() -- hide/show speed display
	:init(0, 0, 0, 0)
	:bordercolor({0.5, 0.5, 0.5, 0.5})
	:movable(false)
	:active(true)
	:clickable(false)
	:visible(true)
	:attach(manoeuvrepanelFrame, 0, 0, mpsize-btnsize/2 , mpheight+htexte-btnsize/2)
	
extracontrolbutton = pCXBox:new() 
	:init(0, 0, 0, 0)
	:bordercolor({0.5, 0.5, 0.5, 0.5})
	:movable(false)
	:active(true)
	:clickable(false)
	:visible(true)
	:attach(manoeuvrepanelFrame, mpsize-btnsize/2, 0, 0, mpheight+htexte-btnsize/2)

control3Drotation = pCXBox:new() 
	:init(0, 0, 0, 0)
	:bordercolor({0.5, 0.5, 0.5, 0.5})
	:fillcolor(controled3Dcolor)
	:movable(false)
	:active(true)
	:clickable(false)
	:visible(true)
	:attach(manoeuvrepanelFrame, btnsize*4-becart, 0, mpsize-btnsize*4+becart - btnsize, mpheight+htexte-btnsize)
	
control3Dtranslation = pCXBox:new()
	:init(0, 0, 0, 0)
	:bordercolor({0.5, 0.5, 0.5, 0.5})
	:fillcolor(controled3Dcolor)
	:movable(false)
	:active(true)
	:clickable(false)
	:visible(true)
	:attach(manoeuvrepanelFrame, btnsize*7+becart, 0, mpsize-btnsize*7-becart - btnsize, mpheight+htexte-btnsize)
	
-- speed display
speedboxheight = 16
trackSpeedDisplay = pCXBox:new()
	:init(0, 0, 0, 0)
	:fillcolor(speedbackground)
	:textfont(normalfont)
	:textcolor(speedontext)
	:movetext(-4, 5)
	:movable(false)
	:active(false)
	:clickable(true)
	:visible(true)
	:attach(screenBox,0,0,width-130,height-speedboxheight)
	
-- extra controls
extracontrolpanel =  pCXBox:new()
	:init(0, 0, 0, 0)
	:bordercolor({0.8, 0.8, 0.8, 0.8})
	:fillcolor({0,0,0,0.8})
	:movable(true)
	:active(false)
	:clickable(false)
	:visible(false)
	:attach(screenBox, expX,expY,width -expX - expsize, height - expY - expheight)

-- cockpits menu

cockpitsmenuFrame = pCXBox:new()
:init(0, 0, 0, 0)
:fillcolor(csetbofill)
:textfont(normalfont)
:textcolor(ctext)
:movetext(ckchoiceMg, 0)
:text(_("Set a cockpit"))
:movable(true)
:clickable(false)
:visible(false)
:attach(screenBox,width-ckchoiceWidth-menuRightpos,height-ckchoiceHeight-2,menuRightpos,2) -- near the edutool panel

btncockpitchoice={}
local createChoiceButton = function(x)
	local separ = 55
	local cockchoice = "< No cockpit >";
	if x > 0 then cockchoice = cockpits[x] end
	return
		pCXBox:new()
        :init(0, 0, 0, 0)
        :textfont(normalfont)
        :textcolor(cbutextoff)
        :movetext(0, 4)
        :text(_(cockchoice))
        :movable(false)
		:active(true)
		:clickable(false)
		:visible(true)
        :attach(cockpitsmenuFrame, ckchoiceMg, ckchoiceHeight-separ-(20*x), 2, separ+(20*x)-20)
end
for i=0, #cockpits do 
	btncockpitchoice[i] = createChoiceButton(i) 
	btncockpitchoice[i].ck = i
end

visorFrame = pCXBox:new()
    :init(0, 0, 0, 0)
    :fillimage(celestia:loadtexture("viseur.png"))
    :movable(false)
    :active(false)
    :clickable(true)
    :visible(false)
    :attach(screenBox, (width-visorsize)/2, (height-visorsize)/2, (width-visorsize)/2, (height-visorsize)/2)

----------------------------------------------
-- local functions
----------------------------------------------
local function round(x,d)
	if x == 0 then return 0 end
	d = d or 0
	local abx = math.abs(x)
	local sign = abx/x
	local m = 10^d
	return math.floor(abx * m +0.5)/m*sign
end

local function sround(x,d)
	if x==0 then return 0 end
	if d then d=d-1 else d=2 end 
	local abx = math.abs(x)
	local sign = abx/x
	return round(abx, -math.floor(math.log10(abx))+d)*sign
end

local function inverserotation(rotation)
	local imag, real = rotation:imag(), rotation:real()
	return celestia:newrotation(-real, imag.x, imag.y, imag.z)
end


local function set3Dposition()
	return cockpit3D and engine and (not engine:visible())
end

local function requirepower()
	acceltime = math.floor(react / perf)
	deltapower = (powercmd-power)/acceltime
end

local initrotation = function(angle,axe)
	if angle == 0 then -- recalculate rotation speed
		if nbrot == math.huge then angle = math.huge end
	end
	if axe ~= axerot then nbrot = 0 end
	axerot = axe
	if angle == math.huge then
		deltarot = rotationspeed/degree * perf
		nbrot = math.huge
	else
		local newangle = angle + nbrot * deltarot
		nbrot = math.floor(newangle*degree / rotationspeed / perf)+1
		deltarot = newangle / nbrot
	end
end

local function initspeed()
	local v0 = ckspeed
	local sel = celestia:getselection()
	local reference = refobj or sel
	if set3Dposition() and ( sel:type() == "spacecraft" or sel:type() == "location") then
		reference = sel
	end
	if reference then
		if set3Dposition() and sel:name() == model3D then
			v0 = 0.0001/km -- 10 cm/s
		else
			local refpos = reference:getposition()
			local radius = reference:radius()
			local dist = obs:getposition():distanceto(refpos) - radius
			local veloc = dist
			if dist <= radius/10 then veloc = radius/10 end
			local N = 50 -- vitesse expo à partir de N*radius
			if dist > radius *N then 
				veloc = dist * (math.log((dist / radius)+1,2) - math.log(N+1,2) + 1)
			end
			v0 = math.max(veloc/km / 10, 0.00001/km) -- mini = 1cm/s
		end
	end
	return v0
end

local adaptzoom = function()
	if zoomcoef then
		fov=2*math.atan(zoomcoef*height/width)
		obs:setfov(fov) -- automatically limited to 120° by celestia
	end
end

local adjustimage = function()
	if cockpitsdata[ck] and cockpitsdata[ck]["image"]  then
		image = cockpitsdata[ck]["image"][panel]
		if image  then
			local X = (image[2] or 0);		if X<1 then X = X * width end
			local Y = (image[3] or 0);		if Y<1 then Y = Y * height end
			local W = (image[4] or 0.1);	if W<1 then W = W * width end
			local H = (image[5] or 0.1);	if H<1 then H = H * height end
			ckpictureFrame:attach(cockpitsFrame,X, Y,width-X-W,height-Y-H)
		end
	end
end

local verifcockpitsize = function() -- check all windows' sizes
	cpX = (width - cpsize)/2	-- recentre control panel
	mpX = cpX+cpsize+becart
	if mpX + mpsize + 6 > width then
		mpX = width - mpsize -3 
		cpX = math.max(mpX - becart - cpsize, 3)
	end
	if mpsize + 2 > width then manoeuvrepanelFrame.Visible = false end
	expX  = mpX + mpsize + becart
	if expX + expsize + 6 > width then expX = width - expsize -3 end
	cpY, mpY, expY = Vmg, Vmg, Vmg
	rotationpanelFrame:attach(screenBox, cpX,cpY,width -cpX - cpsize, height - cpY - cpsize)
	manoeuvrepanelFrame:attach(screenBox, mpX,mpY,width -mpX - mpsize, height - mpY - mpheight-htexte)
	extracontrolpanel:attach(screenBox, expX,expY,width -expX - expsize, height - expY - expheight)
	cockpitsmenuFrame:attach(screenBox,width-ckchoiceWidth-menuRightpos,height-ckchoiceHeight-2,menuRightpos,2)
	
	local mvh, mvv = math.max((width-visorsize)/2,2), math.max((height-visorsize)/2,2)
	visorFrame:attach(screenBox, mvh, mvv, mvh, mvv)
	
	adjustimage()
	adaptzoom()
end

function tourne(axe,angle)
	local rot = celestia:newrotation(axe,angle)
	obs:rotate(rot)
	if vv then
		rot = celestia:newrotation(axe,-angle)
		vv = rot:transform(vv):normalize()
	end
end

local changepanelview = function(way)
	if cockpit3D then return end
	local roth = celestia:newrotation(UP,azim*way)
	local invroth = celestia:newrotation(UP,-azim*way)
	local rotv = celestia:newrotation(DR,elev*way)
	local invrotv = celestia:newrotation(DR,-elev*way)
	
	local vdepl = celestia:newvector(deplX/m, deplY/m, -deplZ/m) * way
	
	local w = vv; if not w then w = oldvv end
	if celvv then w = celvv end
	
	if way == direct then
		local pos = obs:getposition()
		local orient = obs:getorientation()
		local vdir = orient:transform(vdepl)
		local pos2 = pos:addvector(vdir)
		obs:setposition(pos2)
		
		if w then
			obs:rotate(roth);
			w = invroth:transform(w):normalize()
			obs:rotate(rotv)
			w = invrotv:transform(w):normalize()
		end
	else
		if w then
			obs:rotate(rotv);
			w = invrotv:transform(w):normalize()
			obs:rotate(roth)
			w = invroth:transform(w):normalize()
		end
		local pos = obs:getposition()
		local orient = obs:getorientation()
		local vdir = orient:transform(vdepl)
		local pos2 = pos:addvector(vdir)
		obs:setposition(pos2)
	end

	if vv then vv = w elseif oldvv then oldvv = w end
	if celvv then celvv = w end
end

local toggleNavig = function ()
	if manoeuvrepanelFrame.Visible then
		if visorFrame.Visible then visorFrame.Visible = false
			else manoeuvrepanelFrame.Visible = false
		end
	else
		if visorFrame.Visible then 
			manoeuvrepanelFrame.Visible = true
			manoeuvrepanelFrame:orderfront()
		else 
			visorFrame.Visible = true
		end
	end
	verifcockpitsize()
end

local cockpitswitch = function()
	if ck ~= 0 then
		backview = false
		if cockpit3D then
			model:setvisible(not model:visible())
		else
			cockpitsFrame.Visible = not cockpitsFrame.Visible
		end
		if multipanel then 
			rotationpanelFrame.Visible = cockpitsFrame.Visible
		end
	end
end



local loadimage = function() -- try to load [ck,panel]
	if cockpitsdata[ck] and not loading then
		if panel == 0 then 
			panel = cockpitsdata[ck]["firstpanel"] or 1
		end
	else -- loading the entire cockpit
		local name = tostring(cockpits[ck])
		loading = true	
		if panel == 0 then
			panel = 1
			celestia:print(_("Loading cockpit ").._(name).." ...",10)
			return
		end
		if not cockpitsdata[ck] then cockpitsdata[ck] = {} end
		local tx = nil
		local viewname = cockpitdir..name..".png" -- try simple cockpit first
		if panel == 1 then
			tx = celestia:loadtexture(viewname)
			if tx then loading = false end
		end
		if not tx then -- multipanel or panel error
			viewname =  cockpitdir..name.."/"..name.."-"..panel..".png"
			tx = celestia:loadtexture(viewname)
		end
		if tx then -- the panel exists
			cockpitsdata[ck][panel] = {}
			cockpitsdata[ck][panel]["tx"] = tx

		else -- cockpit loading done
			loading = false
			if panel == 1 then 
				if not cockpitsdata[ck][panel] then cockpitsdata[ck][panel]={} end
				cockpitsdata[ck][panel]["tx"]=bug
			else
				panel = panel -1
			end
		end
		if loading then
			panel = panel + 1 -- try to load the next panel
		else  -- done build the panel description
			for p=1, panel do
				cockpitsdata[ck][p]["topanel"]={}
				for i=0, 4 do cockpitsdata[ck][p]["topanel"][i] = 0 end
			end
			-- reads the panel description if any
			local paneldata = _G[cockpits[ck]]
			if not paneldata then 
				if pcall( function() require (cockpits[ck]) end ) then
					paneldata = _G[cockpits[ck]]
				end
			end
			if paneldata then
				panelratio, cockpitfov = paneldata["panelratio"], paneldata["vfov"]
				rearpanel = paneldata["rearpanel"]
				frontpanel = paneldata["frontpanel"] or 1
				if frontpanel > panel then frontpanel = 1 end
				local firstpanel = tonumber(paneldata["firstpanel"]) or 1
				if firstpanel > panel then firstpanel = 1 end
				cockpitsdata[ck]["cockpit3D"] = paneldata["cockpit3D"]
				cockpitsdata[ck]["model3D"] = paneldata["model3D"]
				cockpitsdata[ck]["SolarSystem"] = paneldata["SolarSystem"]
				cockpitsdata[ck]["firstpanel"] = firstpanel
				cockpitsdata[ck]["rearpanel"] = tonumber(rearpanel) or 0
				cockpitsdata[ck]["frontpanel"] = frontpanel
				cockpitsdata[ck]["image"] = paneldata["image"]
				cockpitsdata[ck]["fov3D"] = paneldata["fov"]
				if panelratio and cockpitfov then 
					cockpitsdata[ck]["cfov"] = cockpitfov
				end
				if paneldata["orientation"] then -- special orientation
					for p=1, panel do
						cockpitsdata[ck][p]["azim"] =paneldata["orientation"][p][1]/degree
						cockpitsdata[ck][p]["elev"] =paneldata["orientation"][p][2]/degree
						cockpitsdata[ck][p]["vf"] =paneldata["orientation"][p][3]
					end
				end
				if paneldata["displacement"] then -- special displacement
					for p=1, panel do
						if paneldata["displacement"][p] then 
							cockpitsdata[ck][p]["deplX"] =paneldata["displacement"][p][1]
							cockpitsdata[ck][p]["deplY"] =paneldata["displacement"][p][2]
							cockpitsdata[ck][p]["deplZ"] =paneldata["displacement"][p][3]
						end
					end
				end
				
				if paneldata["viewpath"] then
					for p=1, panel do
						if paneldata["viewpath"][p][5] then
							cockpitsdata[ck][p]["topanel"][0] =paneldata["viewpath"][p][5]
						else
							cockpitsdata[ck][p]["topanel"][0]=frontpanel
						end
						for arrow=1,4 do
							cockpitsdata[ck][p]["topanel"][arrow] =paneldata["viewpath"][p][arrow]
						end
					end
				end
			end
			if panel >1 and cockpitsdata[ck][panel]["topanel"][0] == 0 then
				local to
				for p=1, panel do
					cockpitsdata[ck][p]["topanel"][0] = 1
					to= p-1; if to==0 then to = panel end
					cockpitsdata[ck][p]["topanel"][1] = to
					to= p+1; if to > panel then to = 1 end
					cockpitsdata[ck][p]["topanel"][2] = to
				end
			end
			panel = cockpitsdata[ck]["firstpanel"] or 1
			celestia:print("")
		end
	end
	if not loading then -- parameters of the cockpit
		zoomcoef = nil
		nbpanel = # cockpitsdata[ck]
		cockpit3D = cockpitsdata[ck]["cockpit3D"]
		if cockpit3D then
			SolarSystem = cockpitsdata[ck]["SolarSystem"]
			model3D = cockpitsdata[ck]["model3D"]
			engine = celestia:find("Sol/"..model3D.."_translation_gear")
			steering = celestia:find("Sol/"..model3D.."_rotation_gear")
			model = celestia:find(SolarSystem.."/"..model3D)
			model:setvisible(true)
			cockpitsFrame.Fillimage = nil
			local fov3D=cockpitsdata[ck]["fov3D"]
			if fov3D then obs:setfov(fov3D / degree) end
			if not engine:visible() then
				celestia:select(model)
			end
		else
			cockpitsFrame:fillimage(cockpitsdata[ck][panel]["tx"])
		end

		rearpanel = cockpitsdata[ck]["rearpanel"] or 0
		if rearpanel > nbpanel then rearpanel = 0 end
		frontpanel = cockpitsdata[ck]["frontpanel"] or 1
		if frontpanel > nbpanel then frontpanel = 1 end
		multipanel = (nbpanel > 1)
		for i=0,4 do btncockpitrotFrame[i].panel=0 end
		pictTex = nil
			
		if multipanel then
			local anglerot = 2*math.pi / nbpanel -- default rotation
			azim = cockpitsdata[ck][panel]["azim"] or  anglerot*(panel-1)
			elev = cockpitsdata[ck][panel]["elev"] or 0
			local vf = cockpitsdata[ck][panel]["vf"] or cockpitsdata[ck]["cfov"]
			if vf and panelratio then zoomcoef = panelratio * math.tan(vf/360*math.pi) end
			deplX = cockpitsdata[ck][panel]["deplX"] or 0
			deplY = cockpitsdata[ck][panel]["deplY"] or 0
			deplZ = cockpitsdata[ck][panel]["deplZ"] or 0
			for a=0,4 do btncockpitrotFrame[a].panel = cockpitsdata[ck][panel]["topanel"][a] end
		end
			
		local image = cockpitsdata[ck]["image"]
		if image and image[panel] then
			if not image[panel]["texture"] then
				local pictName = cockpitdir..tostring(cockpits[ck]).."/"..image[panel][1]
				cockpitsdata[ck]["image"][panel]["texture"]=celestia:loadtexture(pictName)
			end
			pictTex = image[panel]["texture"]
			ckpictureFrame:fillimage(pictTex)
			adjustimage()
		end
		
		adaptzoom()
		if newpanel then
			changepanelview(direct)
			newpanel = false
		end
	end
end

local togglecockpit = function()
	cockpitsmenuFrame.Visible = not cockpitsmenuFrame.Visible
	if cockpitsmenuFrame.Visible then 
		cockpitsmenuFrame:orderfront() 
		if lua_edu_tools then
			cockpitsCheck.Bordercolor=cbubordon
			cockpitsCheck.Fillcolor = {cbubordon[1]*2/3, cbubordon[2]*2/3, cbubordon[3]*2/3, cbubordon[4]/2}
		end
	else
		if lua_edu_tools then
			cockpitsCheck.Bordercolor=cbubordoff
			cockpitsCheck.Fillcolor = nil
		end
	end
end

local exit3Dcockpit = function()
	if cockpit3D  then
		cockpit3D = false
		if engine:visible() or steering:visible() then
			engine:setvisible(true)
			steering:setvisible(true)
			engine:setradius(1)
			model:setvisible(false)
		else
			if cockpitvector then
				celestia:print(_"Can't leave ".._(cockpits[ck]).."\n".. _"stop it before",10)
				cockpit3D = true
			else
				powercmd = cockpitsdata[ck]["extpwcmd"]
				cockpitsdata[ck]["intpwcmd"] = 1
			end
		end
	end
end

local nocockpit = function()
	cockpitsFrame.Visible = false;
	cockpitsFrame.Fillimage=nil
	rotationpanelFrame.Visible = false
	multipanel = false
	exit3Dcockpit()
end

local activateCockpit = function(this) -- called by the choice menu
	togglecockpit()
	exit3Dcockpit()
	if not cockpit3D then 
		cockpitsFrame.Visible = true
		backview = false
		ck=this.ck
		panel = 0
		azim, elev = 0,0
		newpanel = true
		loadimage()
		verifcockpitsize()
	end
end

local togglerot = function()
	if nbrot == 0 and oldrotangle then
		initrotation(oldrotangle,axerot)
		oldrotangle = 0
	else
		oldrotangle = nbrot*deltarot
		nbrot = 0
	end
end
		
local function doaccel(vector)
	obs:setspeed(0)
	if not vv then
		power = 0
		if multipanel then changepanelview(inverse) end
		vv= vector; powercmd=math.abs(powercmd)
		if remainingdistance == math.huge then remainingdistance = 0 end
		remainingdistance = remainingdistance + distance
		if multipanel then changepanelview(direct) end
	else
		powercmd = powercmd*impulse
	end
	requirepower()
end

local function accel() doaccel(AV) end

local function reverse()
		if celestiaspeed ~=0 then 
			obs:setspeed(-obs:getspeed())
		elseif not vv then 
			doaccel(AR)
		else
			powercmd = - powercmd
			requirepower()
			powercmd = - powercmd
		end
end


local function slow()
	if refobj then
		powercmd=powercmd/impulse
		requirepower()
	else
		ckspeed = ckspeed/impulse
	end
end

local function stop()
	obs:setspeed(0)
	power = 0
	if vv then
		vv = nil
	else
		vv = oldvv
		power=powercmd
	end
end

local stopall = function()
	if vv or nbrot ~=0 then
		if vv 
			then stop() 
			else oldvv = nil
		end
		if nbrot  ~=0
			then togglerot()
			else oldrotangle = 0
		end
	else
		stop()
		togglerot()
	end
end

local function gostraight()
	if steering then steering:setvisible(true) end 
	if multipanel then
		if cockpitsmenuFrame.Visible then togglecockpit() end
		panel=1
		loadimage()
	end
	power = 0
	if vv then vv = AV end
	powercmd=math.abs(powercmd)
	requirepower()
end

local function lookback()
	obs:rotate(celestia:newrotation(UP, math.pi))
	if vv then vv = vv*-1 end
end

------------------
-- boxes functions
------------------

btncockpitchoice[0].Action = function()
	nocockpit()
	togglecockpit()
end
for i=1, #cockpits do btncockpitchoice[i].Action = activateCockpit end

local verifposition = function()
	obs=celestia:getobserver()
	local obspos = obs:getposition()
	local orient = obs:getorientation()
	if cockpit3D and steering:visible() then 
		cockpitorientation = orient 
	end
	local sc = celestia:getscripttime()
	local dt = sc - scripttime
	perf = (perf*29+dt)/30
	scripttime = celestia:getscripttime()
	frame = obs:getframe()
	refobj = frame:getrefobject()
	celestiaspeed = obs:getspeed()
	if celestiaspeed~=0 then
		vv=nil oldvv=nil 
		if not celvv then celvv = AV end
	else
		celvv = nil
	end

	if acceltime>0 then
		power = power + deltapower
		if power<0 then
			power =-power; deltapower = -deltapower
			if vv then vv=vv*-1 end
		end
		acceltime = acceltime -1
	end
	
	if vv then
		oldvv = vv
		if refobj or set3Dposition() then ckspeed = initspeed() * power end
		local v =  ckspeed * dt
		if remainingdistance <= v then v = remainingdistance end
		if extracontrolpanel.Visible then 
			local dvv = vv:normalize() * v * km
			dx = dx + dvv:getx()
			dy = dy + dvv:gety()
			dz = dz + dvv:getz()
		end
		local vdir
		vdir = orient:transform(vv) * v 
		obspos = obspos:addvector(vdir) -- new position
		obs:setposition(obspos)
		remainingdistance = remainingdistance - v
		if remainingdistance <= 0 then stop() end
		if aheadposition then
			aheadposition = obspos:addvector(vdir*100)
		end
	else
		if celestiaspeed == 0 then
			if refobj or set3Dposition() then ckspeed = initspeed() * powercmd end
		else
			ckspeed=celestiaspeed
		end
	end
	
	-- track management
	if (tracked or aheadposition) and vv then
		local trackpos
		if tracked then
			trackpos = tracked:getposition()
		else
			trackpos = aheadposition
		end
		local neworient = obspos:orientationto(trackpos, orient:transform(UP))
		neworient = orient:slerp(neworient,fracrot)
		obs:setorientation(neworient)
		if fracrot == 1 then
			aheadposition = nil
		else
			fracrot = math.min(fracrot+perf/2,1)
		end
		local vdir = orient:transform(vv):normalize()
		local invrot = inverserotation(neworient)
		vv = invrot:transform(vdir):normalize()
	end
	
	if cockpit3D and aheadposition and not vv then
		local neworient = cockpitorientation
		neworient = orient:slerp(neworient,fracrot)
		obs:setorientation(neworient)
		if fracrot == 1 then
			aheadposition = nil
		else
			fracrot = math.min(fracrot+perf/2,1)
		end
	end
	
	if tracked 
		then speedboxheight = 32
		else speedboxheight = 16
	end
	trackSpeedDisplay:attach(screenBox,0, 0,width-130,height-speedboxheight)

	
	-- rotation management
	if nbrot >0 then
		tourne(axerot, deltarot)
		nbrot = nbrot -1
		if extracontrolpanel.Visible then 
			if         axerot == GA then drotX = drotX - deltarot*degree
				elseif axerot == DR then drotX = drotX + deltarot*degree
				elseif axerot == UP then drotY = drotY + deltarot*degree
				elseif axerot == DN then drotY = drotY - deltarot*degree
				elseif axerot == AV then drotZ = drotZ - deltarot*degree
				elseif axerot == AR then drotZ = drotZ + deltarot*degree
			end
		end
	end

	-- extra control reinit
	if not extracontrolpanel.Visible then 
		dx=0; dy=0; dz=0;
		drotX = 0; drotY = 0; drotZ = 0;
	end

end

trackSpeedDisplay.Customdraw = function(this)
	-- display tracked name if any
	_textlayout:setfont(normalfont)
	_textlayout:setlinespacing(lspace)
	_textlayout:setpos(this.lb+5,this.tb-lspace*1+2) -- position of the first line
	if tracked then
		_textlayout:setfontcolor({0.25,0.5,1,1})
		_textlayout:println("> "..getlocalname(tracked))
	end

	-- speed display
	local v= math.abs(ckspeed*km*1000)
	local u = _("m/s")
	if v>=1000  then 
		v=v/1000   u=_("km/s")
		if v>=10000 then 
			v=v/km_c   u=_("c")
			if v>=100   then 
				v=v/c_au   u=_("au/s")
				if v>=1000  then 
					v=v/au_ly  u=_("ly/s" )
				end
			end
		end
	end
	v= sround(v,3)
	if celestiaspeed ~= 0 then
		_textlayout:setfontcolor(celestiacolortext)
		_textlayout:println(_("Speed: ")..v.." "..u)
	else
		if vv then
			if set3Dposition() and cockpitvector
				then _textlayout:setfontcolor({0,1,1,1}) -- cyan: moving in a moving cockpit
				else _textlayout:setfontcolor(speedontext)
			end
			_textlayout:println(_("Speed: ")..v.." "..u)
		else
			if set3Dposition() and cockpitvector
				then _textlayout:setfontcolor({1,1,0,1}) -- yellow: cockpit is moving
				else _textlayout:setfontcolor(speedofftext)
			end
			_textlayout:println(_("Speed: ").."("..v.." "..u..")")
		end
	end
end

cockpitsFrame.Customdraw = function() 
	ckpictureFrame.Visible = pictTex
end

nominalpowerbutton.Customdraw = function(this)
	this:textfont(normalfont)
	this:textcolor(ctext)
	this:movetext(0, 5)
	if vv 
		then this:fillcolor({0,1,0,0.25}) -- ON
		else this:fillcolor({1,0,0,0.25}) -- OFF
	end
	if refobj then
		this:text(tostring(sround(math.abs(powercmd),3)))
	else
		this:text("- - -") 
	end
end

distanceunitbutton.Customdraw = function(this)
	this.Text = tabunitdist[unitidx][1]
end

fenangle.Customdraw = function(this)
	if not this.Highlight then
		if  cockpit3D and string.sub(this.Text,1,1) == ">" then
			local cmd = string.sub(this.Text,2)
			if cmd == "0" then
				engine:setradius(1)
				this.Text = oldangleText
			elseif cmd == "000" then
				engine:setradius(2)
				this.Text = oldangleText
			end
		end
		defaultrot= math.abs(tonumber(this.Text) or math.huge)
		if defaultrot == math.huge then
			this.Text = ">>>"
		else
			this.Text = tostring(defaultrot)
		end
	else
		vv = savevv
		obs:setspeed(savecelspeed)
	end
end

fenspeedrot.Customdraw = function(this)
	if nbrot>0 then
		this:fillcolor(rotspeedcoloron)
	else
		this:fillcolor(rotspeedcoloroff)
	end
end

distancebox.Customdraw = function(this)
	local dist= math.abs(tonumber(this.Text) or math.huge)
	distance = dist * tabunitdist[unitidx][2]
	if not this.Highlight then
		if distance == math.huge then
			this.Text = ">>>"
		else
			this.Text = tostring(dist)
		end
	else
		vv = savevv
		obs:setspeed(savecelspeed)
	end
end

local depldistance = function(x)
	local u = " km"
	local d = x
	if math.abs(x) <1 then
		d=d*1000
		u= " m"
	elseif math.abs(x) > 9.46e12 then
		d=d/km/10^6
		u = " al"
	end
	if u == " m" then
		return round(d,2)..u
	else
		return sround(d,5)..u
	end
end

extracontrolpanel.Customdraw = function(this)
	_textlayout:setfont(normalfont)
	_textlayout:setfontcolor(ctext)
	_textlayout:setlinespacing(lspace)
	_textlayout:setpos(this.lb+5,this.tb-lspace*1+2) -- position of the first line
	_textlayout:println("dx = "..depldistance(dx))
	_textlayout:println("dy = "..depldistance(dy))
	_textlayout:println("dz = "..depldistance(-dz))
	_textlayout:println("")
	_textlayout:println("drotY = "..round(drotY,2))
	_textlayout:println("drotX = "..round(drotX,2))
	_textlayout:println("drotZ = "..round(drotZ,2))
	
	if refobj then
		local obspos = obs:getposition()
		local refpos = refobj:getposition()
		local vecteur = obspos:vectorto(refpos)
		local orient = obs:getorientation()
		local invrot = inverserotation(orient)
		local vecdepl = invrot:transform(vecteur) * km
		_textlayout:println("")
		_textlayout:println("px = "..depldistance( (vecdepl.x) ))
		_textlayout:println("py = "..depldistance( (vecdepl.y) ))
		_textlayout:println("pz = "..depldistance( -(vecdepl.z) ))
	end
end

control3Drotation.Customdraw = function(this)
	if steering then 
		if steering:visible() then
			this:fillcolor(controled3Dcolor)
		else
			this:fillcolor(free3Dcolor)
		end
	end
end

control3Dtranslation.Customdraw = function(this)
	if engine then 
		if engine:visible() then
			this:fillcolor(controled3Dcolor)
		else
			this:fillcolor(free3Dcolor)
		end
	end
end

--
-- ACTIONS
--

control3Drotation.Action = function(this)
	if steering then steering:setvisible(not steering:visible()) end
end

control3Dtranslation.Action = function(this)
	if engine then 
		engine:setvisible(not engine:visible())
		if set3Dposition() then 
			oldcockpit3Dselection = celestia:getselection()
			cockpitvector = nil
			cockpitsdata[ck]["cockpitframe"] = obs:getframe()
			cockpitsdata[ck]["extpwcmd"] = powercmd
			powercmd = cockpitsdata[ck]["intpwcmd"] or 1
			if vv then
				cockpitvector = obs:getorientation():transform(vv)
				cockpitspeed = ckspeed * perf
				vv= nil
				power = 0
			end
			celestia:select(model)
		else
			vv = nil
			if cockpitvector then
				vv = inverserotation(obs:getorientation()):transform(cockpitvector)
			end
			cockpitsdata[ck]["intpwcmd"] = powercmd
			powercmd = cockpitsdata[ck]["extpwcmd"]
			power = powercmd
			obs:setframe(cockpitsdata[ck]["cockpitframe"])
			if celestia:getselection():name() == model3D then celestia:select(cockpitsdata[ck]["cockpitframe"]:getrefobject()) end
		end
	end
	oldvv = nil
end

local btnrotAction = function(this,pan)
	local oldpan = panel
	changepanelview(inverse)
	if pan then
		panel = pan
	else
		if this.panel == frontpanel and oldpan==frontpanel and rearpanel~=0 then
			panel = rearpanel
		else
			panel = this.panel
		end
	end
	if cockpitsmenuFrame.Visible then togglecockpit() end
	loadimage()
	changepanelview(direct)
end
for i=0,4 do btncockpitrotFrame[i].Action = btnrotAction end

speeddisplaybutton.Action = function() 
	displayspeed = not displayspeed 
end

extracontrolbutton.Action = function()
	showextra = not showextra
end

	
fenangle.Action = function(this)
	oldangleText = this.Text
	nbrot = 0
	this.Text = ""
	savevv = vv
	savecelspeed = celestiaspeed
end

distancebox.Action = function(this)
	remainingdistance = 0
	this.Text = ""
	savevv = vv
	savecelspeed = celestiaspeed
end

decspeedrot.Action = function(this)
	rotationspeed = rotationspeed / rotimpulse
	fenspeedrot.Text = tostring(round(rotationspeed))
	initrotation(0,axerot)
end

incspeedrot.Action = function(this)
	rotationspeed = rotationspeed * rotimpulse
	fenspeedrot.Text = tostring(round(rotationspeed))
	initrotation(0,axerot)
end


if lua_edu_tools then
	cockpitsCheck.Action = function(this)
		togglecockpit()
	end
end


local dirAction = function(this) 
	obs:setspeed(0)
	local btn = this.btn
	if btn==0 or btn==7 then stop()
	elseif btn < 8 then -- translation
		if refobj then
			power = 0
			powercmd=math.abs(powercmd)
			requirepower()	
		end
		vv = direction[btn]
		
		if remainingdistance == math.huge then remainingdistance = 0 end
		remainingdistance = remainingdistance + distance
		
	elseif btn ~= 14 then -- rotations
		initrotation(defaultrot/degree, direction[btn])
	else
		togglerot()
	end
end
for i = 0,14 do btncockpitmanFrame[i].Action = dirAction end

slowbutton.Action = slow

accelbutton.Action = function() 
	if refobj then
		powercmd=powercmd*impulse
		requirepower()
	else
		ckspeed = ckspeed*impulse
	end
end

power1xbutton.Action = function() 
	if refobj then
		powercmd = powercmd / math.abs(powercmd)
		requirepower()
	end
end

distanceunitbutton.Action = function(this)
	unitidx = unitidx + 1
	if unitidx > #tabunitdist then unitidx = 1 end
end

-- program main loop

cockpitsSurvey.Customdraw = function(this)
	verifposition()
	oldwidth, oldheight = width, height
	width, height = celestia:getscreendimension()
	trackSpeedDisplay.Visible = displayspeed
	extracontrolpanel.Visible = showextra and manoeuvrepanelFrame.Visible 
	control3Drotation.Visible = cockpit3D
	control3Dtranslation.Visible = cockpit3D -- and (not vv) and (obs:getspeed()== 0)
	if cockpit3D then
		local sel = celestia:getselection()
		if engine:visible() then
			if sel:name() == model3D then
				celestia:select(oldcockpit3Dselection)
			else
				oldcockpit3Dselection = sel
			end
		elseif sel:name() ~= model3D then
			oldcockpit3Dselection = sel
		end
	end
	local textoverlay = celestia:getoverlayelements()
	textoverlay.Velocity = false
	celestia:setoverlayelements(textoverlay)
		
	local changescreendim = (oldwidth ~= width or oldheight ~= height)
	if cockpitsFrame.Visible then
		if loading then loadimage() end
		obs=celestia:getobserver()
		local cpvisible = false
		for i=0,4 do 
			btncockpitrotFrame[i].Visible = (btncockpitrotFrame[i].panel >0)
			cpvisible = cpvisible or btncockpitrotFrame[i].Visible
		end
		rotationpanelFrame.Visible = cpvisible
	end
	if changescreendim then 
		verifcockpitsize()
	end
end

local Track=function() 
	if tracked then
		tracked = nil
	else
		tracked = celestia:getselection()
		if tracked:name() == "?" then
			tracked = nil
		end
	end
	if tracked then
		fracrot = perf/2
		aheadposition = nil
	end
end

local Viewahead=function()
	if vv or (cockpit3D and not steering:visible()) then
		tracked = nil
		aheadposition = true
		fracrot = perf/2
	end
end

----------------------------------------------------

keymap["w"] = accel
keymap["z"] = slow
keymap["s"] = stopall
keymap["q"] = reverse
keymap["x"] = gostraight
keymap["*"] = lookback

keymap[manoeuvre_track_key or "T"] = Track
keymap[manoeuvre_view_ahead_key or "R"] = Viewahead

local shortcut = cockpits_shortcut or "Z"
keymap[shortcut] = togglecockpit

local key = visor_shorcut or "X"
keymap[key] = toggleNavig

if cockpit_switch_key then  -- no default key
	keymap[cockpit_switch_key] = cockpitswitch
end

celestia:log("-- cockpitsBox V3D loaded")